use quote::quote;
use syn::{
    Ident, Lit, Token, Type, braced,
    parse::{Parse, ParseStream},
    punctuated::Punctuated,
    token,
};

#[proc_macro]
pub fn schema(item: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let schema: SchemaDescription = syn::parse(item).unwrap();
    generate_structure(schema, None).into()
}

#[derive(Debug)]
struct SchemaDescription {
    name: Ident,
    value: Option<(token::Colon, Type)>,
    fields: Option<(token::Brace, Punctuated<SchemaField, Token![,]>)>,
}
impl Parse for SchemaDescription {
    fn parse(input: ParseStream) -> syn::Result<Self> {
        let content;
        let name = input.parse()?;
        let value = if input.peek(token::Colon) {
            Some((Parse::parse(input)?, Parse::parse(input)?))
        } else {
            None
        };
        let fields = if input.peek(token::Brace) {
            Some((
                braced!(content in input),
                content.parse_terminated(Parse::parse, Token![,])?,
            ))
        } else {
            None
        };
        Ok(Self {
            name,
            value,
            fields,
        })
    }
}

#[derive(Debug)]
enum SchemaKey {
    Literal(Lit),
    Type(Type),
}
impl Parse for SchemaKey {
    fn parse(input: ParseStream) -> syn::Result<Self> {
        if input.peek(Lit) {
            input.parse().map(Self::Literal)
        } else {
            input.parse().map(Self::Type)
        }
    }
}

#[derive(Debug)]
struct SchemaField {
    key: SchemaKey,
    #[allow(unused)]
    colon_token: Token![:],
    subschema: SchemaDescription,
}
impl Parse for SchemaField {
    fn parse(input: ParseStream) -> syn::Result<Self> {
        Ok(Self {
            key: input.parse()?,
            colon_token: input.parse()?,
            subschema: input.parse()?,
        })
    }
}

fn generate_structure(
    schema: SchemaDescription,
    path: Option<(&Ident, SchemaKey)>,
) -> proc_macro2::TokenStream {
    let SchemaDescription {
        name,
        value,
        fields,
    } = schema;
    let constructor = match path {
        None => quote! {
            impl ::std::convert::From<::foundationdb::directory::DirectorySubspace> for #name {
                fn from(space: ::foundationdb::directory::DirectorySubspace) -> Self {
                    Self(::foundationdb::tuple::Subspace::from_bytes(space.bytes().to_owned()))
                }
            }
        },
        Some((parent, SchemaKey::Literal(Lit::Str(key)))) => {
            let key_ident = syn::parse_str::<Ident>(&key.value()).ok().into_iter();
            quote! {
                impl ::std::convert::From<&#parent> for #name {
                    fn from(space: &#parent) -> Self {
                        Self(space.0.subspace(&#key))
                    }
                }
                #(
                    impl #parent {
                        pub fn #key_ident(&self) -> #name {
                            self.into()
                        }
                    }
                )*
            }
        }
        Some((parent, SchemaKey::Literal(key))) => quote! {
            impl ::std::convert::From<&#parent> for #name {
                fn from(space: &#parent) -> Self {
                    Self(space.0.subspace(&#key))
                }
            }
        },
        Some((parent, SchemaKey::Type(key_type))) => quote! {
            impl <T: ::std::borrow::Borrow<#key_type>> ::std::ops::Div<&T> for &#parent {
                type Output = #name;
                fn div(self, key: &T) -> #name {
                    #name(self.0.subspace(&::std::borrow::Borrow::borrow(key)))
                }
            }
            impl <T: ::std::borrow::Borrow<#key_type>> ::std::ops::Div<&T> for #parent {
                type Output = #name;
                fn div(self, key: &T) -> #name {
                    #name(self.0.subspace(&::std::borrow::Borrow::borrow(key)))
                }
            }
        },
    };
    let children = fields.into_iter().flat_map(|(_, fields)| {
        fields.into_iter().map(
            |SchemaField {
                 key,
                 colon_token: _,
                 subschema,
             }| { generate_structure(subschema, Some((&name, key))) },
        )
    });
    let value = value.into_iter().map(|(_, value_type)| {
        quote! {
            impl ::fdb_schema::FdbTypedKey for #name {
                type Representation = #value_type;
                fn get_raw_key(&self) -> &[u8] { self.0.bytes() }
            }
        }
    });
    quote! {
        pub struct #name(::foundationdb::tuple::Subspace);

        impl ::std::ops::Deref for #name {
            type Target = ::foundationdb::tuple::Subspace;
            fn deref(&self) -> &Self::Target { &self.0 }
        }

        #constructor
        #(#value)*
        #(#children)*
    }
}
